/*
  Solar Router V2.01
  ******************

  Measure Voltage, Current and Power

  In case of overproduction of solar electricity, activate a Triac and a relay(optional) 
  to connect a resistive load able to absorb the overproduction.
  
  Details and hardware wiring are available in :
  https://f1atb.fr  in the Home Automation / Domotique section

  F1ATB November 2022

  GNU General Public License v3.0

*/

//Libraries
#include <WiFi.h>
#include <ESPmDNS.h>
#include <WiFiUdp.h>
#include <WiFiClient.h>
#include <WebServer.h>
#include <ArduinoOTA.h>
#include <RemoteDebug.h>
#include <NTPClient.h>

//Program routines
#include "PageWeb.h"

//WIFI
const char* ssid = "SSID"; //Put here your WIFI SSID
const char* password = "xxxxxxx"; //Put here the WIFI password
// Set your Static IP address
IPAddress local_IP(192, 168, 0, 208);
// Set your Gateway IP address
IPAddress gateway(192, 168, 0, 254);
IPAddress subnet(255, 255, 255, 0);
IPAddress primaryDNS(8, 8, 8, 8);    //optional
IPAddress secondaryDNS(8, 8, 4, 4);  //optional

//Debug via WIFI instead of Serial
//Connect a Telnet terminal on port 23
RemoteDebug Debug;


//PINS - GPIO
const int AnalogIn0 = 35;
const int AnalogIn1 = 32;
const int AnalogIn2 = 33;
const int zeroCross = 23;
const int pulseTriac = 22;
const int LedYellow = 18;
const int LedGreen = 19;
const int OptionalRelay = 5;


//VARIABLES
int value0;
int volt[100];
int amp[100];
int sinus[100];  //Table of sinus
int ModeHeater = 0;
int WIFIbug = 0;
int LastHour = -1;

float Uef;  //Voltage
float Ief;  //Current
float PVA;  //Power in VA
float PW;   //Power in Watt
float PowerFactor;
float kV = 0.2083;    //Calibration coefficient for the voltage. To be adjusted
float kI = 0.0642;    //Calibration coefficient for the current. To be adjusted
float Wh = 0;         //Watt*hour
float THDv;           // Total Harmonic Distortion including Noise of the voltage
float THDi;           // Total Harmonic Distortion including Noise of the current
float voltM[100];     //Voltage Mean value
float ampM[100];      //Amps Mean value
float retardF = 100;  //Floating value of retard(delay)

//Internal Timers
unsigned long previousComputeMillis;
unsigned long previousBlinkMillis;
unsigned long previousWifiMillis;

bool Blink = false;

//Optional - Client of Domoticz
const char* host = "192.168.0.99";
const int httpPort = 8080;
const int idxPower = 1012;



//Variables in RAM for interruptions
volatile unsigned long lastIT = 0;
volatile int delayTriac = 0;
volatile int retard = 100;
hw_timer_t* timer = NULL;
volatile bool phasePositive = true;
volatile int Idx0 = 0;  //Ref Storage index

WebServer server(80);  // Simple Web Server on port 80


WiFiUDP ntpUDP;
NTPClient temps(ntpUDP, "fr.pool.ntp.org", 3600, 300000);  // External timeb  (Udp,pool,offset in second 3600=winter in europe,update in ms)
//  SERVER
//***********
void handleRoot() {  //Main web page called
  Debug.println(F("Client Web"));
  String S = String(getPageWeb());
  server.send(200, "text/html", S);
}
void handleAjaxData() {  //Periodic call to refresh the data on the web page
  String S = String(Uef) + "," + String(Ief) + "," + String(PW) + "," + String(PVA) + "," + String(PowerFactor) + ",";
  S += String(Wh / 1000) + "," + String(floor(millis() / 36000) / 100) + "," + String(ModeHeater) + "," + String(retard) + ",";
  S += String(int(THDv)) + "," + String(int(THDi)) + "|";
  for (int i = 0; i < 100; i++) {
    S += String(int(10 * voltM[i])) + ",";  //Voltages*10. Increase dynamic
  }
  S += "0|";
  for (int i = 0; i < 100; i++) {
    S += String(int(10 * ampM[i])) + ",";  //Currents*10
  }
  S += "0";

  server.send(200, "text/html", S);  // All data grouped in a large serial stream
}
void handleAjaxOrder() {  // Order received to force the opening of the Triac
  ModeHeater = server.arg(0).toInt();
  retard = min(100, 125 - 25 * ModeHeater);  //100 Triac not used. 0 Triac open 100% of the time
  retardF = float(retard);
  server.send(200, "text/html", String(ModeHeater));
}

void handleRestart() {  // Order received to force the opening of the Triac
  ESP.restart();
}

void handleNotFound() {  //Page not found
  Debug.println(F("File Not Found"));
  String message = "File Not Found\n\n";
  message += "URI: ";
  message += server.uri();
  message += "\nMethod: ";
  message += (server.method() == HTTP_GET) ? "GET" : "POST";
  message += "\nArguments: ";
  message += server.args();
  message += "\n";
  for (uint8_t i = 0; i < server.args(); i++) {
    message += " " + server.argName(i) + ": " + server.arg(i) + "\n";
  }
  server.send(404, "text/plain", message);
}

// DOMOTICZ client - Optional to transfer the measurement to a home automation system
//****************
void SendToDomoticz() {
  String url;
  // Use WiFiClient class to create TCP connections
  WiFiClient client;
  if (!client.connect(host, httpPort)) {
    Serial.println("connection failed");
    return;
  }
  // We now create a URI for the request
  url = "/json.htm?type=command&param=udevice&idx=";
  url += String(idxPower);
  url += "&nvalue=0&svalue=";
  url += String(PW);
  url += ";";
  url += String(Wh);

  // This will send the request to the Domoticz server
  client.print(String("GET ") + url + " HTTP/1.1\r\n" + "Host: " + host + "\r\n" + "Connection: close\r\n\r\n");
  unsigned long timeout = millis();
  while (client.available() == 0) {
    if (millis() - timeout > 5000) {
      Serial.println(">>> Client Timeout !");
      client.stop();
      return;
    }
  }

  // Read all the lines of the reply from server and print them to Serial
  while (client.available()) {
    String line = client.readStringUntil('\r');
    // Serial.print(line); //No interest of the response
  }
}

//Interruption, Current Zero Crossing from Triac device
void IRAM_ATTR currentNull() {
  if ((millis() - lastIT) > 2) {  // to avoid glitch detection during 2ms
    delayTriac = 0;               //Time synchro every 10ms
    lastIT = millis();
    digitalWrite(pulseTriac, LOW);  //Stop Triac
    phasePositive = !phasePositive;
    if (phasePositive ){
      Idx0 =(micros() % 20000) / 200;  // To synchronise the voltage and curent short integration and the curves to display     
    }
    
  }
}
// Interruption Timer
void IRAM_ATTR onTimer() {  //Interruption every 100 micro second
  delayTriac += 1;
  if (delayTriac > retard && retard < 98) {  //100 steps in 10 ms
    digitalWrite(pulseTriac, HIGH);          //Activate Triac
  }
}
// POWER
//********
void MeasurePower() {
  int iStore;
  value0 = analogRead(AnalogIn0);  //Mean value. Should be at 3.3v/2
  unsigned long MeasureMillis = millis();
  
  while (millis() - MeasureMillis < 21) {               //Read values in continuous during 20ms. One loop is around 150 micro seconds
    iStore = ((micros() % 20000) / 200+100 - Idx0) % 100;  //We have more results that we need during 20ms to fill the tables of 100 samples
    volt[iStore] = analogRead(AnalogIn1) - value0;
    amp[iStore] = analogRead(AnalogIn2) - value0;
  }
}
void ComputePower() {
  float V;
  float V1 = 0;
  float V2 = 0;
  float Uef2 = 0;
  float I;
  float I1 = 0;
  float I2 = 0;
  float Ief2 = 0;
  Uef = 0;
  Ief = 0;
  PW = 0;
  for (int i = 0; i < 100; i++) {
    voltM[i] = (19 * voltM[i] + float(volt[i])) / 20;  //Mean value. First Order Filter. Short Integration
    V = kV * voltM[i];
    Uef += sq(V);
    V1 += V * float(sinus[i]);                      //Sine
    V2 += V * float(sinus[(i + 25) % 100]);         //Cosine
    ampM[i] = (19 * ampM[i] + float(amp[i])) / 20;  //Mean value. First Order Filter
    I = kI * ampM[i];
    Ief += sq(I);
    I1 += I * float(sinus[i]);
    I2 += I * float(sinus[(i + 25) % 100]);  //Cosinus
    PW += V * I;
  }
  Uef2 = Uef / 100;  //square of voltage
  Uef = sqrt(Uef2);  //RMS voltage
  Ief2 = Ief / 100;  //square of current
  Ief = sqrt(Ief2);  // RMS current
  PW = PW / 100;
  PVA = floor(Uef * Ief);
  PowerFactor = floor(100 * PW / PVA) / 100;
  Wh += PW / 90000;                                   // Watt Hour,Every 40ms
  PW = floor(PW );
  THDv = 2 * (sq(V1 / 3000000) + sq(V2 / 3000000));  //(Level of Voltage Fondemental frequency 50Hz)²
  if (Uef2 >= THDv && Uef > 0) {
    THDv = floor(100 * sqrt(Uef2 - THDv) / Uef);  // Total Harmonic Distorsion + Noise for the Voltage
  } else {
    THDv = 0;
  }
  THDi = 2 * (sq(I1 / 3000000) + sq(I2 / 3000000));  //(Level of Current Fondemental frequency 50 Hz)²
  if (Ief2 >= THDi && Ief > 0) {
    THDi = floor(100 * sqrt(Ief2 - THDi) / Ief);  // Total Harmonic Distorsion + Noise for the Current
  } else {
    THDi = 0;
  }
}

void Overproduction() {
  if (ModeHeater == 0) {           //Water Heater / Chauffe-Eau automatique
                                   // Overproduction
    retardF = retardF + PW / 200;  //Decrease/Increase delay to increase/Decrease Triac conduction
    if (retardF < 0) { retardF = 0; }
    if (retardF > 100) { retardF = 100; }
    retard = int(retardF);  //integer value used to drive the triac
  }
  //Optional Relay. To avoid oscillations, the difference between the 2 thresholds On and Off
  // MUST be bigger than the power consumed by the device
  if (PW < -1200) {  //switch On a relay for any  device
    digitalWrite(OptionalRelay, HIGH);
  }
  if (PW > 0) {  //Switch Off
    digitalWrite(OptionalRelay, LOW);
  }
}

void Read_Hour() {
  //Time Update / de l'heure
  temps.update();
  Serial.println(temps.getFormattedTime());
  int hour = int(temps.getHours());
  if (LastHour != 0 && hour == 0) { Wh = 0; }  // Reset Watt Hour of the day. Once every day
  if (hour >= 23 || hour <= 4) {               //Force water Heater On during the night. Values OK for Summer and Winter even if one hour error
    ModeHeater = 5;
    retard = 0;
    retardF = 0;
  }
  if (LastHour == 4 && hour == 5) { ModeHeater = 0; }  // Back to Automatic mode
  LastHour = hour;
}


// SETUP
//*******
void setup() {
  pinMode(LedYellow, OUTPUT);
  pinMode(LedGreen, OUTPUT);
  pinMode(zeroCross, INPUT);
  pinMode(pulseTriac, OUTPUT);
  pinMode(OptionalRelay, OUTPUT);
  digitalWrite(LedYellow, LOW);
  digitalWrite(LedGreen, LOW);
  digitalWrite(pulseTriac, LOW);     //Stop Triac
  digitalWrite(OptionalRelay, LOW);  //Stop Relay

  Serial.begin(115200);
  Serial.println("Booting");
  for (int i = 0; i < 100; i++) {  //Reset table measurements
    voltM[i] = 0;
    ampM[i] = 0;
  }
  // Configures static IP address
  if (!WiFi.config(local_IP, gateway, subnet, primaryDNS, secondaryDNS)) {
    Serial.println("STA Failed to configure");
  }
  //WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);
  while (WiFi.waitForConnectResult() != WL_CONNECTED) {
    Debug.println("Connection Failed! Rebooting...");
    delay(5000);
    ESP.restart();
  }
  // init remote debug
  Debug.begin("ESP32");  // Telnet on port 23
  //Init Software update via WIFI (On The Air)
  initOTA();

  Debug.println("Ready");
  Debug.print("IP address: ");
  Debug.println(WiFi.localIP());
  Serial.print("IP address: ");
  Serial.println(WiFi.localIP());

  //Init Web Server on port 80
  server.on("/", handleRoot);
  server.on("/ajax_data", handleAjaxData);
  server.on("/ajax_order", handleAjaxOrder);
  server.on("/restart", handleRestart);
  server.onNotFound(handleNotFound);
  server.begin();
  Debug.println("HTTP server started");

  //Various Timers
  previousComputeMillis = millis();
  previousBlinkMillis = millis();
  previousWifiMillis = millis();



  attachInterrupt(zeroCross, currentNull, RISING);

  //Hardware timer
  timer = timerBegin(0, 80, true);  //Clock Divider, 1 micro second Tick
  timerAttachInterrupt(timer, &onTimer, true);
  timerAlarmWrite(timer, 100, true);  //Interrupt every 100 Ticks or microsecond
  timerAlarmEnable(timer);

  //Table of Sinus, to speed up THD computation
  for (int angle = 0; angle <= 100; angle++) {
    float angle_rad = angle * M_PI / 50;
    sinus[angle] = int(30000 * sin(angle_rad));
  }

  //External timer to obtain the Hour and reset Watt Hour every day at 0h
  temps.begin();
}

// LOOP
//******
void loop() {
  ArduinoOTA.handle();
  Debug.handle();
  server.handleClient();

  if (millis() - previousComputeMillis >= 40) {
    previousComputeMillis = millis();
    MeasurePower();
    ComputePower();
    Overproduction();
    
  }
  if (millis() - previousBlinkMillis >= 2000) {
    Blink = !Blink;
    if (Blink) {
      previousBlinkMillis = millis() - 1950;
    } else {
      previousBlinkMillis = millis();
    }
    if (PW >= 0) {
      digitalWrite(LedYellow, Blink);
      digitalWrite(LedGreen, LOW);
    } else {
      digitalWrite(LedYellow, LOW);
      digitalWrite(LedGreen, Blink);
    }
    Debug.println("V25:" + String(voltM[25]));
    Debug.println("Idx0:" + String(Idx0));
    
  }
  if (millis() - previousWifiMillis > 30000) {
    previousWifiMillis = millis();
    if (WiFi.waitForConnectResult() != WL_CONNECTED) {
      Debug.println("Connection Failed! #" + String(WIFIbug));
      WIFIbug++;
      if (WIFIbug > 20) {
        ESP.restart();
      }
    } else {
      WIFIbug = 0;
    }
    
    Read_Hour();
    Debug.println("Power:" + String(PW));
    Serial.println("Power:" + String(PW));
    if (voltM[25] <0){ // Should be positive for the display
      phasePositive = !phasePositive;
    }
    SendToDomoticz();  //Optional to inform Home Automation system
  }
}

// OTA (code from Arduino example)
//******
void initOTA() {
  // Port defaults to 3232
  // ArduinoOTA.setPort(3232);
  // Hostname defaults to esp3232-[MAC]
  //ArduinoOTA.setHostname("ESP32-Power-V2");
  // No authentication by default
  //ArduinoOTA.setPassword("admin");
  // Password can be set with it's md5 value as well
  // MD5(admin) = 21232f297a57a5a743894a0e4a801fc3
  // ArduinoOTA.setPasswordHash("21232f297a57a5a743894a0e4a801fc3");
  ArduinoOTA
    .onStart([]() {
      String type;
      if (ArduinoOTA.getCommand() == U_FLASH)
        type = "sketch";
      else  // U_SPIFFS
        type = "filesystem";
      // NOTE: if updating SPIFFS this would be the place to unmount SPIFFS using SPIFFS.end()
      Debug.println("Start updating " + type);
    })
    .onEnd([]() {
      Debug.println("\nEnd");
    })
    .onProgress([](unsigned int progress, unsigned int total) {
      Debug.printf("Progress: %u%%\r", (progress / (total / 100)));
    })
    .onError([](ota_error_t error) {
      Debug.printf("Error[%u]: ", error);
      if (error == OTA_AUTH_ERROR) Debug.println("Auth Failed");
      else if (error == OTA_BEGIN_ERROR) Debug.println("Begin Failed");
      else if (error == OTA_CONNECT_ERROR) Debug.println("Connect Failed");
      else if (error == OTA_RECEIVE_ERROR) Debug.println("Receive Failed");
      else if (error == OTA_END_ERROR) Debug.println("End Failed");
    });
  ArduinoOTA.begin();
}
